// Shader2sidedView.cpp : implementation of the CShader2sidedView class
//

#include "stdafx.h"
#include "Shader2sided.h"
#include "Shader2sidedDoc.h"
#include "Shader2sidedView.h"

#include "dx9/RendererDX9.h"
#include "Misc/MyAssert.h"

#include <mmsystem.h>

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// CShader2sidedView

IMPLEMENT_DYNCREATE(CShader2sidedView, CView)

BEGIN_MESSAGE_MAP(CShader2sidedView, CView)
	//{{AFX_MSG_MAP(CShader2sidedView)
	ON_WM_ERASEBKGND()
	ON_COMMAND(ID_2SIDED, On2sided)
	ON_UPDATE_COMMAND_UI(ID_2SIDED, OnUpdate2sided)
	ON_COMMAND(ID_WIREFRAME, OnWireframe)
	ON_UPDATE_COMMAND_UI(ID_WIREFRAME, OnUpdateWireframe)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CShader2sidedView construction/destruction

CShader2sidedView::CShader2sidedView()
{
  m_bo2Sided    = true;
  m_boWireframe = false;
}

CShader2sidedView::~CShader2sidedView()
{
}

BOOL CShader2sidedView::PreCreateWindow(CREATESTRUCT& cs)
{
	// TODO: Modify the Window class or styles here by modifying
	//  the CREATESTRUCT cs

	return CView::PreCreateWindow(cs);
}

/////////////////////////////////////////////////////////////////////////////
// CShader2sidedView drawing

void CShader2sidedView::OnDraw(CDC* pDC)
{
	CShader2sidedDoc* pDoc = GetDocument();
	ASSERT_VALID(pDoc);
	// TODO: add draw code for native data here
}

/////////////////////////////////////////////////////////////////////////////
// CShader2sidedView diagnostics

#ifdef _DEBUG
void CShader2sidedView::AssertValid() const
{
	CView::AssertValid();
}

void CShader2sidedView::Dump(CDumpContext& dc) const
{
	CView::Dump(dc);
}

CShader2sidedDoc* CShader2sidedView::GetDocument() // non-debug version is inline
{
	ASSERT(m_pDocument->IsKindOf(RUNTIME_CLASS(CShader2sidedDoc)));
	return (CShader2sidedDoc*)m_pDocument;
}
#endif //_DEBUG

/////////////////////////////////////////////////////////////////////////////
// CShader2sidedView message handlers

BOOL CShader2sidedView::OnEraseBkgnd(CDC* pDC) 
  {
  if(DrawFrame()) return TRUE;
	return CView::OnEraseBkgnd(pDC);
  }

//

void CShader2sidedView::OnUpdate(CView* pSender, LPARAM lHint, CObject* pHint) 
  {
	switch(lHint)
    {
    case CShader2sidedApp::_IDLE_ : DrawFrame(); break;
    }
  }

//

bool CShader2sidedView::DrawFrame()
  {
  Mythos::CRendererDX9* pRenderer = theApp.m_pRenderer;
  if(!pRenderer) return false;

  // camera rotates around (0;0;0)

  static float fAngle = 0.f;

  Mythos::CMat4x4 m4Trans,m4Rot,m4Trf;
  m4Trans.Translation(Mythos::CVect3D(0.f,0.f,10.f));
  m4Rot.RotationY(fAngle);
  m4Trf = m4Rot*m4Trans;
  pRenderer->SetWorld2View(m4Trf.RotTransInverted());

  static DWORD dwLastTime = timeGetTime();
  DWORD        dwNewTime  = timeGetTime();
  DWORD        dwTimeDiff = dwNewTime-dwLastTime;
  dwLastTime = dwNewTime;

  fAngle += float(dwTimeDiff)/1000.f*2.f*_PI_/4.f;
  while(fAngle > 2*_PI_) fAngle -= 2*_PI_;

  // draw

  _ASSERT_(pRenderer->BeginScene(),"BeginScene failed");
  _ASSERT_(pRenderer->Clear(),     "Clear failed");

  pRenderer->Set2Sided   (m_bo2Sided);
  pRenderer->SetWireframe(m_boWireframe);
  pRenderer->Draw();

  _ASSERT_(pRenderer->EndScene(),"EndScene failed");
  _ASSERT_(pRenderer->Swap(),    "Swap failed");
  return true;
  }

//

void CShader2sidedView::On2sided() 
  {
  m_bo2Sided = !m_bo2Sided;
  }

void CShader2sidedView::OnUpdate2sided(CCmdUI* pCmdUI) 
  {
  pCmdUI->SetCheck(m_bo2Sided);
  }

void CShader2sidedView::OnWireframe() 
  {
  m_boWireframe = !m_boWireframe;
  }

void CShader2sidedView::OnUpdateWireframe(CCmdUI* pCmdUI) 
  {
  pCmdUI->SetCheck(m_boWireframe);
  }
