//--------------------------------------------------------------------------------------------------------------------//
//                                                DX9 RENDERER CLASS                                                  //
//--------------------------------------------------------------------------------------------------------------------//

//----------------------------------------------- INCLUDES -----------------------------------------------------------//

#include  "stdafx.h"
#include  "RendererDX9.h"
#ifdef    _DEBUG
#include  "RendererDX9.inl"
#endif

//----------------------------------------------- STATIC MEMBERS -----------------------------------------------------//

namespace Mythos
{

//--------------------------------------------------------------------------------------------------------------------//
//                                                BASIC OPS                                                           //
//--------------------------------------------------------------------------------------------------------------------//

//----------------------------------------------- Create -------------------------------------------------------------//
// create & init renderer
// 20/05/02, Mythos
// in : window handle,flags(=_WINDOWED_|_ZBUFFER_),width(=0) & height(=0)
// out: OK?
// rem: if width==height==0 the client area of the window is used
//--------------------------------------------------------------------------------------------------------------------//

bool CRendererDX9::Create(const HWND hWnd,const u32 u32Flags,const u32 u32Width,const u32 u32Height)
  {
  Destroy();

  bool boWindowed =(u32Flags & _WINDOWED_) != 0;
  bool boZBuffer  =(u32Flags & _ZBUFFER_ ) != 0;
  u32  u32Width2  = u32Width;
  u32  u32Height2 = u32Height;

  if(boWindowed)
    {
    RECT rect;
    if(!GetClientRect(hWnd,&rect)) return false;
    if(!u32Width)  u32Width2  = rect.right;
    if(!u32Height) u32Height2 = rect.bottom;
    }

  //

  m_pD3D = Direct3DCreate9(D3D_SDK_VERSION);
  if(!m_pD3D) return false;

  D3DDISPLAYMODE d3ddm;
  if(FAILED(m_pD3D->GetAdapterDisplayMode(D3DADAPTER_DEFAULT,&d3ddm))) return false;

  DWORD dwVtxProc = D3DCREATE_SOFTWARE_VERTEXPROCESSING;
  D3DCAPS9 d3dcaps;
  if(SUCCEEDED(m_pD3D->GetDeviceCaps(D3DADAPTER_DEFAULT,D3DDEVTYPE_HAL,&d3dcaps)))
    {
    if(d3dcaps.DevCaps & D3DDEVCAPS_HWTRANSFORMANDLIGHT) dwVtxProc = D3DCREATE_HARDWARE_VERTEXPROCESSING;
    }

  D3DPRESENT_PARAMETERS d3dpp; 
  ZeroMemory(&d3dpp,sizeof(d3dpp));
  d3dpp.BackBufferWidth             = u32Width2;
  d3dpp.BackBufferHeight            = u32Height2;
  d3dpp.BackBufferFormat            = d3ddm.Format;
  d3dpp.BackBufferCount             = 1;
  d3dpp.MultiSampleType             = D3DMULTISAMPLE_NONE;
  d3dpp.MultiSampleQuality          = 0;
  d3dpp.SwapEffect                  = D3DSWAPEFFECT_DISCARD;
  d3dpp.hDeviceWindow               = hWnd;
  d3dpp.Windowed                    = boWindowed;
  d3dpp.EnableAutoDepthStencil      = boZBuffer;
  d3dpp.AutoDepthStencilFormat      = D3DFMT_D16;
  d3dpp.Flags                       = 0;
  d3dpp.FullScreen_RefreshRateInHz  = 0;
  d3dpp.PresentationInterval        = D3DPRESENT_INTERVAL_IMMEDIATE;

  HRESULT hrErr = m_pD3D->CreateDevice(D3DADAPTER_DEFAULT,D3DDEVTYPE_HAL,hWnd,dwVtxProc,&d3dpp,&m_pDevice);
  if(FAILED(hrErr))
    {
    TCHAR* pszError = (TCHAR*)DXGetErrorString9(hrErr);
    _ASSERT_(false,"CreateDevice failed");
    _ASSERT_(false,pszError);
    return false;
    }

  D3DCAPS9 d3dDevDesc9;
  if(FAILED(m_pDevice->GetDeviceCaps(&d3dDevDesc9))) return false;

  //

  D3DVIEWPORT9 d3dViewport;
  d3dViewport.X      = 0;
  d3dViewport.Y      = 0;
  d3dViewport.Width  = u32Width2;
  d3dViewport.Height = u32Height2;
  d3dViewport.MinZ   = 0.f;
  d3dViewport.MaxZ   = 1.f;
  m_pDevice->SetViewport(&d3dViewport);

  InitStates();
  if(!InitScene()) return false;

  // trfs

  SetModel2World(Mythos::CMat4x4::Identity);
  SetWorld2View (Mythos::CMat4x4::Identity);
  SetPerspective(_DEG2RAD_(60.f),1.f,1000.f);

  return true;
  }

//----------------------------------------------- InitStates ---------------------------------------------------------//
// init render & texture stages states
// 24/05/02, Mythos
// in :
// out:
//--------------------------------------------------------------------------------------------------------------------//

void CRendererDX9::InitStates()
  {
  // render states

  float fZero = 0.f;
  float fOne  = 1.f;
  float f64   = 64.f;

  bool boLighted = false;
  bool boZTest   = true;
  bool boCulling = true;

  D3DCOLOR colTexFactor   = 0xff80ff80;
  D3DCOLOR colBlendFactor = 0x80808080;

  m_pDevice->SetRenderState(D3DRS_FILLMODE,                 D3DFILL_SOLID);
  m_pDevice->SetRenderState(D3DRS_SHADEMODE,                D3DSHADE_GOURAUD);
  m_pDevice->SetRenderState(D3DRS_LASTPIXEL,                true);
  m_pDevice->SetRenderState(D3DRS_CULLMODE,                 boCulling ? D3DCULL_CW : D3DCULL_NONE);
  m_pDevice->SetRenderState(D3DRS_DITHERENABLE,             true);
  m_pDevice->SetRenderState(D3DRS_CLIPPING,                 true);
  m_pDevice->SetRenderState(D3DRS_VERTEXBLEND,              D3DVBF_DISABLE);
  m_pDevice->SetRenderState(D3DRS_CLIPPLANEENABLE,          0);
  m_pDevice->SetRenderState(D3DRS_DEBUGMONITORTOKEN,        D3DDMT_ENABLE);
  m_pDevice->SetRenderState(D3DRS_INDEXEDVERTEXBLENDENABLE, false);
  m_pDevice->SetRenderState(D3DRS_COLORWRITEENABLE,         0xf);
  m_pDevice->SetRenderState(D3DRS_COLORWRITEENABLE1,        0xf);
  m_pDevice->SetRenderState(D3DRS_COLORWRITEENABLE2,        0xf);
  m_pDevice->SetRenderState(D3DRS_COLORWRITEENABLE3,        0xf);
  m_pDevice->SetRenderState(D3DRS_SRGBWRITEENABLE,          0);
  m_pDevice->SetRenderState(D3DRS_TWEENFACTOR,              *(DWORD*)&fZero);
  m_pDevice->SetRenderState(D3DRS_BLENDOP,                  D3DBLENDOP_ADD);

  m_pDevice->SetRenderState(D3DRS_ANTIALIASEDLINEENABLE,    false);
  m_pDevice->SetRenderState(D3DRS_MULTISAMPLEANTIALIAS,     false);
  m_pDevice->SetRenderState(D3DRS_MULTISAMPLEMASK,          0xffffffff);

  m_pDevice->SetRenderState(D3DRS_PATCHEDGESTYLE,           D3DPATCHEDGE_DISCRETE);
  m_pDevice->SetRenderState(D3DRS_POSITIONDEGREE,           D3DDEGREE_CUBIC);
  m_pDevice->SetRenderState(D3DRS_NORMALDEGREE,             D3DDEGREE_LINEAR);

  m_pDevice->SetRenderState(D3DRS_MINTESSELLATIONLEVEL,     *(DWORD*)&fOne);
  m_pDevice->SetRenderState(D3DRS_MAXTESSELLATIONLEVEL,     *(DWORD*)&fOne);
  m_pDevice->SetRenderState(D3DRS_ADAPTIVETESS_X,           *(DWORD*)&fZero);
  m_pDevice->SetRenderState(D3DRS_ADAPTIVETESS_Y,           *(DWORD*)&fZero);
  m_pDevice->SetRenderState(D3DRS_ADAPTIVETESS_Z,           *(DWORD*)&fOne);
  m_pDevice->SetRenderState(D3DRS_ADAPTIVETESS_W,           *(DWORD*)&fZero);

  m_pDevice->SetRenderState(D3DRS_ZENABLE,                  boZTest);
  m_pDevice->SetRenderState(D3DRS_ZWRITEENABLE,             true);
  m_pDevice->SetRenderState(D3DRS_ZFUNC,                    D3DCMP_LESSEQUAL);

  m_pDevice->SetRenderState(D3DRS_ALPHATESTENABLE,          false);
  m_pDevice->SetRenderState(D3DRS_SRCBLEND,                 /*D3DBLEND_ONE*/D3DBLEND_SRCALPHA);
  m_pDevice->SetRenderState(D3DRS_DESTBLEND,                /*D3DBLEND_ZERO*/D3DBLEND_INVSRCALPHA);
  m_pDevice->SetRenderState(D3DRS_ALPHAREF,                 0);
  m_pDevice->SetRenderState(D3DRS_ALPHAFUNC,                D3DCMP_ALWAYS);
  m_pDevice->SetRenderState(D3DRS_ALPHABLENDENABLE,         true);

  m_pDevice->SetRenderState(D3DRS_SEPARATEALPHABLENDENABLE, false);
  m_pDevice->SetRenderState(D3DRS_SRCBLENDALPHA,            D3DBLEND_ONE);
  m_pDevice->SetRenderState(D3DRS_DESTBLENDALPHA,           D3DBLEND_ZERO);
  m_pDevice->SetRenderState(D3DRS_BLENDOPALPHA,             D3DBLENDOP_ADD);

  m_pDevice->SetRenderState(D3DRS_FOGENABLE,                false);
  m_pDevice->SetRenderState(D3DRS_FOGCOLOR,                 0);
  m_pDevice->SetRenderState(D3DRS_FOGTABLEMODE,             D3DFOG_NONE);
  m_pDevice->SetRenderState(D3DRS_FOGSTART,                 *(DWORD*)&fZero);
  m_pDevice->SetRenderState(D3DRS_FOGEND,                   *(DWORD*)&fZero);
  m_pDevice->SetRenderState(D3DRS_FOGDENSITY,               *(DWORD*)&fOne);
  m_pDevice->SetRenderState(D3DRS_RANGEFOGENABLE,           false);
  m_pDevice->SetRenderState(D3DRS_FOGVERTEXMODE,            D3DFOG_NONE);

  m_pDevice->SetRenderState(D3DRS_SPECULARENABLE,           false);
  m_pDevice->SetRenderState(D3DRS_LIGHTING,                 boLighted);
  m_pDevice->SetRenderState(D3DRS_AMBIENT,                  D3DCOLOR_XRGB(128,128,128));
  m_pDevice->SetRenderState(D3DRS_COLORVERTEX,              false);
  m_pDevice->SetRenderState(D3DRS_LOCALVIEWER,              true);
  m_pDevice->SetRenderState(D3DRS_NORMALIZENORMALS,         false);
  m_pDevice->SetRenderState(D3DRS_DIFFUSEMATERIALSOURCE,    D3DMCS_MATERIAL);
  m_pDevice->SetRenderState(D3DRS_SPECULARMATERIALSOURCE,   D3DMCS_MATERIAL);
  m_pDevice->SetRenderState(D3DRS_AMBIENTMATERIALSOURCE,    D3DMCS_MATERIAL);
  m_pDevice->SetRenderState(D3DRS_EMISSIVEMATERIALSOURCE,   D3DMCS_MATERIAL);

  m_pDevice->SetRenderState(D3DRS_STENCILENABLE,            false);
  m_pDevice->SetRenderState(D3DRS_STENCILFAIL,              D3DSTENCILOP_KEEP);
  m_pDevice->SetRenderState(D3DRS_STENCILZFAIL,             D3DSTENCILOP_KEEP);
  m_pDevice->SetRenderState(D3DRS_STENCILPASS,              D3DSTENCILOP_KEEP);
  m_pDevice->SetRenderState(D3DRS_STENCILFUNC,              D3DCMP_ALWAYS);
  m_pDevice->SetRenderState(D3DRS_STENCILREF,               0);
  m_pDevice->SetRenderState(D3DRS_STENCILMASK,              0xffffffff);
  m_pDevice->SetRenderState(D3DRS_STENCILWRITEMASK,         0xffffffff);

  m_pDevice->SetRenderState(D3DRS_TWOSIDEDSTENCILMODE,      false);
  m_pDevice->SetRenderState(D3DRS_CCW_STENCILFAIL,          D3DSTENCILOP_KEEP);
  m_pDevice->SetRenderState(D3DRS_CCW_STENCILZFAIL,         D3DSTENCILOP_KEEP);
  m_pDevice->SetRenderState(D3DRS_CCW_STENCILPASS,          D3DSTENCILOP_KEEP);
  m_pDevice->SetRenderState(D3DRS_CCW_STENCILFUNC,          D3DCMP_ALWAYS);

  m_pDevice->SetRenderState(D3DRS_TEXTUREFACTOR,            colTexFactor);
  m_pDevice->SetRenderState(D3DRS_BLENDFACTOR,              colBlendFactor);
  m_pDevice->SetRenderState(D3DRS_WRAP0,                    0);//D3DWRAPCOORD_0 | D3DWRAPCOORD_1 | D3DWRAPCOORD_2 | D3DWRAPCOORD_3);
  m_pDevice->SetRenderState(D3DRS_WRAP1,                    D3DWRAPCOORD_0 | D3DWRAPCOORD_1 | D3DWRAPCOORD_2 | D3DWRAPCOORD_3);
  m_pDevice->SetRenderState(D3DRS_WRAP2,                    D3DWRAPCOORD_0 | D3DWRAPCOORD_1 | D3DWRAPCOORD_2 | D3DWRAPCOORD_3);
  m_pDevice->SetRenderState(D3DRS_WRAP3,                    D3DWRAPCOORD_0 | D3DWRAPCOORD_1 | D3DWRAPCOORD_2 | D3DWRAPCOORD_3);
  m_pDevice->SetRenderState(D3DRS_WRAP4,                    D3DWRAPCOORD_0 | D3DWRAPCOORD_1 | D3DWRAPCOORD_2 | D3DWRAPCOORD_3);
  m_pDevice->SetRenderState(D3DRS_WRAP5,                    D3DWRAPCOORD_0 | D3DWRAPCOORD_1 | D3DWRAPCOORD_2 | D3DWRAPCOORD_3);
  m_pDevice->SetRenderState(D3DRS_WRAP6,                    D3DWRAPCOORD_0 | D3DWRAPCOORD_1 | D3DWRAPCOORD_2 | D3DWRAPCOORD_3);
  m_pDevice->SetRenderState(D3DRS_WRAP7,                    D3DWRAPCOORD_0 | D3DWRAPCOORD_1 | D3DWRAPCOORD_2 | D3DWRAPCOORD_3);

  m_pDevice->SetRenderState(D3DRS_WRAP8,                    D3DWRAPCOORD_0 | D3DWRAPCOORD_1 | D3DWRAPCOORD_2 | D3DWRAPCOORD_3);
  m_pDevice->SetRenderState(D3DRS_WRAP9,                    D3DWRAPCOORD_0 | D3DWRAPCOORD_1 | D3DWRAPCOORD_2 | D3DWRAPCOORD_3);
  m_pDevice->SetRenderState(D3DRS_WRAP10,                   D3DWRAPCOORD_0 | D3DWRAPCOORD_1 | D3DWRAPCOORD_2 | D3DWRAPCOORD_3);
  m_pDevice->SetRenderState(D3DRS_WRAP11,                   D3DWRAPCOORD_0 | D3DWRAPCOORD_1 | D3DWRAPCOORD_2 | D3DWRAPCOORD_3);
  m_pDevice->SetRenderState(D3DRS_WRAP12,                   D3DWRAPCOORD_0 | D3DWRAPCOORD_1 | D3DWRAPCOORD_2 | D3DWRAPCOORD_3);
  m_pDevice->SetRenderState(D3DRS_WRAP13,                   D3DWRAPCOORD_0 | D3DWRAPCOORD_1 | D3DWRAPCOORD_2 | D3DWRAPCOORD_3);
  m_pDevice->SetRenderState(D3DRS_WRAP14,                   D3DWRAPCOORD_0 | D3DWRAPCOORD_1 | D3DWRAPCOORD_2 | D3DWRAPCOORD_3);
  m_pDevice->SetRenderState(D3DRS_WRAP15,                   D3DWRAPCOORD_0 | D3DWRAPCOORD_1 | D3DWRAPCOORD_2 | D3DWRAPCOORD_3);

  m_pDevice->SetRenderState(D3DRS_POINTSIZE,                *(DWORD*)&fOne);
  m_pDevice->SetRenderState(D3DRS_POINTSIZE_MIN,            *(DWORD*)&fZero);
  m_pDevice->SetRenderState(D3DRS_POINTSPRITEENABLE,        false);
  m_pDevice->SetRenderState(D3DRS_POINTSCALEENABLE,         false);
  m_pDevice->SetRenderState(D3DRS_POINTSCALE_A,             *(DWORD*)&fOne);
  m_pDevice->SetRenderState(D3DRS_POINTSCALE_B,             *(DWORD*)&fZero);
  m_pDevice->SetRenderState(D3DRS_POINTSCALE_C,             *(DWORD*)&fZero);
  m_pDevice->SetRenderState(D3DRS_POINTSIZE_MAX,            *(DWORD*)&f64);

  m_pDevice->SetRenderState(D3DRS_SCISSORTESTENABLE,        false);
  m_pDevice->SetRenderState(D3DRS_SLOPESCALEDEPTHBIAS,      0);
  m_pDevice->SetRenderState(D3DRS_DEPTHBIAS,                0);

  // texture stages states

  m_pDevice->SetSamplerState(0, D3DSAMP_MAGFILTER, D3DTEXF_LINEAR);
  m_pDevice->SetSamplerState(0, D3DSAMP_MINFILTER, D3DTEXF_LINEAR);
  m_pDevice->SetSamplerState(0, D3DSAMP_MIPFILTER, D3DTEXF_NONE);

  m_pDevice->SetSamplerState(1, D3DSAMP_MAGFILTER, D3DTEXF_LINEAR);
  m_pDevice->SetSamplerState(1, D3DSAMP_MINFILTER, D3DTEXF_LINEAR);
  m_pDevice->SetSamplerState(1, D3DSAMP_MIPFILTER, D3DTEXF_NONE);

  m_pDevice->SetTexture(0,NULL);
  m_pDevice->SetTexture(1,NULL);

  m_pDevice->SetTextureStageState(0, D3DTSS_TEXCOORDINDEX, 0);
  m_pDevice->SetTextureStageState(0, D3DTSS_COLOROP,   D3DTOP_SELECTARG1);
  m_pDevice->SetTextureStageState(0, D3DTSS_COLORARG1, D3DTA_DIFFUSE);
  m_pDevice->SetTextureStageState(0, D3DTSS_COLORARG2, D3DTA_TEXTURE);
  m_pDevice->SetTextureStageState(0, D3DTSS_ALPHAOP,   D3DTOP_SELECTARG1);
  m_pDevice->SetTextureStageState(0, D3DTSS_ALPHAARG1, D3DTA_DIFFUSE);
  m_pDevice->SetTextureStageState(0, D3DTSS_ALPHAARG2, D3DTA_TEXTURE);

  m_pDevice->SetTextureStageState(1, D3DTSS_TEXCOORDINDEX, 1);
  m_pDevice->SetTextureStageState(1, D3DTSS_COLOROP,   D3DTOP_SELECTARG1);
  m_pDevice->SetTextureStageState(1, D3DTSS_COLORARG1, D3DTA_CURRENT);
  m_pDevice->SetTextureStageState(1, D3DTSS_COLORARG2, D3DTA_TEXTURE);
  m_pDevice->SetTextureStageState(1, D3DTSS_ALPHAOP,   D3DTOP_SELECTARG1);
  m_pDevice->SetTextureStageState(1, D3DTSS_ALPHAARG1, D3DTA_CURRENT);
  m_pDevice->SetTextureStageState(1, D3DTSS_ALPHAARG2, D3DTA_TEXTURE);

  m_pDevice->SetTextureStageState(1, D3DTSS_COLOROP, D3DTOP_DISABLE);
  m_pDevice->SetTextureStageState(1, D3DTSS_ALPHAOP, D3DTOP_DISABLE);

  m_pDevice->SetTextureStageState(2, D3DTSS_COLOROP, D3DTOP_DISABLE);
  m_pDevice->SetTextureStageState(2, D3DTSS_ALPHAOP, D3DTOP_DISABLE);
  }

//----------------------------------------------- InitScene ----------------------------------------------------------//
// init buffers & shaders
// 23/08/03, Mythos
// in :
// out: OK ?
//--------------------------------------------------------------------------------------------------------------------//

bool CRendererDX9::InitScene()
  {
  if(!m_pDevice) return false;

  // IB

  u16 u16IDs[6] = { 0,1,2, 0,2,3 };

  D3DINDEXBUFFER_DESC ibDesc;
  memset(&ibDesc,0,sizeof(ibDesc));
  ibDesc.Format = D3DFMT_INDEX16;
  ibDesc.Size   = sizeof(u16IDs);
  ibDesc.Type   = D3DRTYPE_INDEXBUFFER;
  ibDesc.Pool   = D3DPOOL_DEFAULT;
  ibDesc.Usage  = D3DUSAGE_WRITEONLY;

  HRESULT hrErr = m_pDevice->CreateIndexBuffer(ibDesc.Size,ibDesc.Usage,ibDesc.Format,ibDesc.Pool,&m_pIB,NULL);
  if(FAILED(hrErr)) return false;

  u32 u32Flags   = D3DLOCK_NOSYSLOCK;
  u16* pLockedIB = NULL;
  hrErr = m_pIB->Lock(0,sizeof(u16IDs),(void**)&pLockedIB,u32Flags);
  if(FAILED(hrErr)) return false;

  memcpy(pLockedIB,u16IDs,sizeof(u16IDs));
  hrErr = m_pIB->Unlock();
  if(FAILED(hrErr)) return false;

  // VB:  03
  //      12

  CVect3D v3Left (-1.f, 1.f, 0.f);
  CVect3D v3Right( 1.f,-1.f, 0.f);
  CVect3D v3N    ( 0.f, 0.f, 1.f);
  CVect3D v3Pt[4];
  v3Pt[0].Set(v3Left);
  v3Pt[1].Set(v3Left.x,v3Right.y,v3Left.z);
  v3Pt[2].Set(v3Right);
  v3Pt[3].Set(v3Right.x,v3Left.y,v3Right.z);

  CVect2D v2UV0(0.f);
  CVect2D v2UV1(1.f);
  u32 u32FVF = D3DFVF_XYZ | D3DFVF_NORMAL | D3DFVF_TEX1;

  VERTEX_SIMPLE VtxArray[4] = {
    VERTEX_SIMPLE(v3Pt[0], v3N, v2UV0),
    VERTEX_SIMPLE(v3Pt[1], v3N, CVect2D(v2UV0.u,v2UV1.v)),
    VERTEX_SIMPLE(v3Pt[2], v3N, v2UV1),
    VERTEX_SIMPLE(v3Pt[3], v3N, CVect2D(v2UV1.u,v2UV0.v)),
    };

  D3DVERTEXBUFFER_DESC vbDesc;
  memset(&vbDesc,0,sizeof(vbDesc));
  vbDesc.Format = D3DFMT_VERTEXDATA;
  vbDesc.Size   = sizeof(VtxArray);
  vbDesc.Type   = D3DRTYPE_VERTEXBUFFER;
  vbDesc.Pool   = D3DPOOL_DEFAULT;
  vbDesc.Usage  = D3DUSAGE_WRITEONLY;
  vbDesc.FVF    = u32FVF;

  hrErr = m_pDevice->CreateVertexBuffer(vbDesc.Size,vbDesc.Usage,vbDesc.FVF,vbDesc.Pool,&m_pVB,NULL);
  if(FAILED(hrErr)) return false;

  u32Flags = D3DLOCK_NOSYSLOCK;
  void* pLockedVB = NULL;
  hrErr = m_pVB->Lock(0,sizeof(VtxArray),(void**)&pLockedVB,u32Flags);
  if(FAILED(hrErr)) return false;

  memcpy(pLockedVB,VtxArray,sizeof(VtxArray));
  hrErr = m_pVB->Unlock();
  if(FAILED(hrErr)) return false;

  // vshader

  D3DVERTEXELEMENT9 decl[] =
    {
      { 0,  0, D3DDECLTYPE_FLOAT3, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_POSITION, 0 },
      { 0, 12, D3DDECLTYPE_FLOAT3, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_NORMAL,   0 },
      { 0, 24, D3DDECLTYPE_FLOAT2, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_TEXCOORD, 0 },
        D3DDECL_END()
    };

  hrErr = m_pDevice->CreateVertexDeclaration(decl,&m_pVertexDeclaration);
  if(FAILED(hrErr))
    {
    MessageBox(NULL,"CreateVertexDeclaration failed","CRendererDX9::Create",MB_OK|MB_ICONEXCLAMATION);
    return false;
    }

  DWORD dwFlags = 0; 
  dwFlags |= D3DXSHADER_DEBUG;
  LPD3DXBUFFER pCode   = NULL;
  LPD3DXBUFFER pErrors = NULL;
//  hrErr = D3DXAssembleShaderFromFile("dx9/vshader.vsh",NULL,NULL,dwFlags,&pCode,&pErrors);
  hrErr = D3DXCompileShaderFromFile("dx9/vshader.fx",NULL,NULL,"VShade","vs_1_1",dwFlags,&pCode,&pErrors,&m_pVertexConstants);
  if(pErrors)
    {
    char* szErrors = (char*)pErrors->GetBufferPointer();
    pErrors->Release();
    }
  if(FAILED(hrErr)) 
    {
    MessageBox(NULL,"vertex shader creation failed","CRendererDX9::Create",MB_OK|MB_ICONEXCLAMATION);
    return false;
    }

  char* szCode = (char*)pCode->GetBufferPointer();
  hrErr = m_pDevice->CreateVertexShader((DWORD*)pCode->GetBufferPointer(),&m_pVertexShader);
  pCode->Release();
  if(FAILED(hrErr))
    {
    MessageBox(NULL,"CreateVertexShader failed","CRendererDX9::Create",MB_OK|MB_ICONEXCLAMATION);
    return false;
    }

  // pshader

  hrErr = D3DXCompileShaderFromFile("dx9/pshader.fx",NULL,NULL,"PShade","ps_1_1",dwFlags,&pCode,&pErrors,&m_pPixelConstants);
  if(pErrors)
    {
    char* szErrors = (char*)pErrors->GetBufferPointer();
    pErrors->Release();
    }
  if(FAILED(hrErr)) 
    {
    MessageBox(NULL,"pixel shader creation failed","CRendererDX9::Create",MB_OK|MB_ICONEXCLAMATION);
    return false;
    }

  szCode = (char*)pCode->GetBufferPointer();
  hrErr  = m_pDevice->CreatePixelShader((DWORD*)pCode->GetBufferPointer(),&m_pPixelShader);
  pCode->Release();
  if(FAILED(hrErr))
    {
    MessageBox(NULL,"CreatePixelShader failed","CRendererDX9::Create",MB_OK|MB_ICONEXCLAMATION);
    return false;
    }

  // textures

  hrErr = D3DXCreateTextureFromFile(m_pDevice,"dx9/front.jpg",&m_pTexFront);
  if(FAILED(hrErr))
    {
    MessageBox(NULL,"error loading front.jpg","D3DXCreateTextureFromFile",MB_OK|MB_ICONEXCLAMATION);
    return false;
    }

  hrErr = D3DXCreateTextureFromFile(m_pDevice,"dx9/back.jpg",&m_pTexBack);
  if(FAILED(hrErr))
    {
    MessageBox(NULL,"error loading back.jpg","D3DXCreateTextureFromFile",MB_OK|MB_ICONEXCLAMATION);
    return false;
    }

  //

  return true;
  }

//----------------------------------------------- Destroy ------------------------------------------------------------//
// destroy renderer
// 20/05/02, Mythos
// in :
// out:
//--------------------------------------------------------------------------------------------------------------------//

void CRendererDX9::Destroy()
  {
  if(m_pPixelShader)    m_pPixelShader   ->Release();
  if(m_pPixelConstants) m_pPixelConstants->Release();
  m_pPixelShader    = NULL;
  m_pPixelConstants = NULL;

  if(m_pVertexShader)      m_pVertexShader     ->Release();
  if(m_pVertexDeclaration) m_pVertexDeclaration->Release();
  if(m_pVertexConstants)   m_pVertexConstants  ->Release();
  m_pVertexShader      = NULL;
  m_pVertexDeclaration = NULL;
  m_pVertexConstants   = NULL;

  if(m_pIB) m_pIB->Release();
  if(m_pVB) m_pVB->Release();
  m_pIB = NULL;
  m_pVB = NULL;

  if(m_pDevice) m_pDevice->Release();
  if(m_pD3D)    m_pD3D   ->Release();
  m_pDevice = NULL;
  m_pD3D    = NULL;

  if(m_pTexFront) m_pTexFront->Release();
  if(m_pTexBack)  m_pTexBack ->Release();
  m_pTexFront = NULL;
  m_pTexBack  = NULL;
  }

//----------------------------------------------- BeginScene ---------------------------------------------------------//
// begin frame
// 20/05/02, Mythos
// in :
// out: OK?
//--------------------------------------------------------------------------------------------------------------------//

bool CRendererDX9::BeginScene()
  {
  if(!m_pDevice) return false;
  return SUCCEEDED(m_pDevice->BeginScene());
  }

//----------------------------------------------- EndScene -----------------------------------------------------------//
// end frame
// 20/05/02, Mythos
// in :
// out: OK?
//--------------------------------------------------------------------------------------------------------------------//

bool CRendererDX9::EndScene()
  {
  if(!m_pDevice) return false;
  return SUCCEEDED(m_pDevice->EndScene());
  }

//----------------------------------------------- Clear --------------------------------------------------------------//
// clear buffers
// 20/05/02, Mythos
// in : buffers to clear(=IRenderer::_CLEAR_BACKDEPTH_)
// out: OK?
//--------------------------------------------------------------------------------------------------------------------//

bool CRendererDX9::Clear(const u32 u32Flags)
  {
  if(!m_pDevice) return false;
  
  u32 u32Buffers = 0;
  if(u32Flags & _CLEAR_BACK_)    u32Buffers |= D3DCLEAR_TARGET;
  if(u32Flags & _CLEAR_DEPTH_)   u32Buffers |= D3DCLEAR_ZBUFFER;
  if(u32Flags & _CLEAR_STENCIL_) u32Buffers |= D3DCLEAR_STENCIL;

  D3DCOLOR ClearColor = D3DCOLOR_XRGB(128,128,255);

  HRESULT hrErr = m_pDevice->Clear(0,NULL,u32Buffers,ClearColor,1.f,128);
  if(SUCCEEDED(hrErr)) return true;

  TCHAR* pszError = (TCHAR*)DXGetErrorString9(hrErr);
  _ASSERT_(false,"Clear failed");
  _ASSERT_(false,pszError);
  return false;
  }

//----------------------------------------------- Swap ---------------------------------------------------------------//
// swap front & back buffers
// 20/05/02, Mythos
// in : source(=NULL) & dest(=NULL) rects,nb vbls(=0) to wait (TBI)
// out: OK?
//--------------------------------------------------------------------------------------------------------------------//

bool CRendererDX9::Swap(const RECT* pSrcRect,const RECT* pDestRect,const u32 u32Vbls)
  {
  if(!m_pDevice) return false;
  return SUCCEEDED(m_pDevice->Present(pSrcRect,pDestRect,NULL,NULL));

  UNREFERENCED_PARAMETER(u32Vbls);
  }

//----------------------------------------------- SetWireframe -------------------------------------------------------//
// enable/disable wireframe mode
// 22/09/02, Mythos
// in : enable?
// out: OK?
//--------------------------------------------------------------------------------------------------------------------//

bool CRendererDX9::SetWireframe(const bool boON)
  {
  if(!m_pDevice) return false;
  return SUCCEEDED(m_pDevice->SetRenderState(D3DRS_FILLMODE,boON ? D3DFILL_WIREFRAME : D3DFILL_SOLID));
  }

//----------------------------------------------- Set2Sided ----------------------------------------------------------//
// enable/disable 2-sided mode
// 23/08/03, Mythos
// in : enable?
// out:
//--------------------------------------------------------------------------------------------------------------------//

void CRendererDX9::Set2Sided(const bool boON)
  {
  m_bo2Sided = boON;
  }

//--------------------------------------------------------------------------------------------------------------------//
//                                                DRAW                                                                //
//--------------------------------------------------------------------------------------------------------------------//

//----------------------------------------------- Draw ---------------------------------------------------------------//
// render scene
// 23/08/03, Mythos
// in :
// out: 0=OK
// rem: hardcoded version for the sample without materials and the like
//--------------------------------------------------------------------------------------------------------------------//

u32 CRendererDX9::Draw()
  {
  if(!m_pDevice) return -1;

  // shader constants

  CMat4x4 m4World,m4View,m4Proj,m4Total;
  m_pDevice->GetTransform(D3DTS_WORLD,     (D3DMATRIX*)&m4World);
  m_pDevice->GetTransform(D3DTS_VIEW,      (D3DMATRIX*)&m4View);
  m_pDevice->GetTransform(D3DTS_PROJECTION,(D3DMATRIX*)&m4Proj);

    // ASM

  m4Total = m4Proj*m4View*m4World;
  m4Total.Transpose();
  m_pDevice->SetVertexShaderConstantF(10,(float*)&m4Total,4);         // trf matrix

  CVect4D v4NFactor(1.f);
  m_pDevice->SetVertexShaderConstantF(14,(float*)&v4NFactor,1);       // normal factor

  CVect4D v4Ambient(0.25f,0.25f,0.25f,1.f);
  m_pDevice->SetVertexShaderConstantF(15,(float*)&v4Ambient,1);       // ambient color

  CVect4D v4MatDiffuse(1.f,1.f,1.f,1.f);
  m_pDevice->SetVertexShaderConstantF(16,(float*)&v4MatDiffuse,1);    // material diffuse

      // lights

  CVect4D v4LightDiffuse1(0.f,0.f,1.f,1.f);
  m_pDevice->SetVertexShaderConstantF(20,(float*)&v4LightDiffuse1,1); // light1 diffuse

  m4World.Invert();
  CVect4D v4LightDir1(0.f,0.f,-1.f,0.f);
  v4LightDir1 = m4World*v4LightDir1;
  v4LightDir1.Normalize1();
  m_pDevice->SetVertexShaderConstantF(21,(float*)&v4LightDir1,1);     // light1 direction

  CVect4D v4LightDiffuse2(1.f,0.f,0.f,1.f);
  m_pDevice->SetVertexShaderConstantF(30,(float*)&v4LightDiffuse2,1); // light2 diffuse

  CVect4D v4LightDir2(0.f,0.f,1.f,0.f);
  v4LightDir2 = m4World*v4LightDir2;
  v4LightDir2.Normalize1();
  m_pDevice->SetVertexShaderConstantF(31,(float*)&v4LightDir2,1);     // light2 direction

    // HLSL

  UINT TextureIndex = 0;
  if(m_pVertexConstants)
    {
    D3DXHANDLE handle;
    if(handle = m_pVertexConstants->GetConstantByName(NULL,"Matrix"))
      m_pVertexConstants->SetMatrix(m_pDevice,handle,(D3DXMATRIX*)&m4Total);

    if(handle = m_pVertexConstants->GetConstantByName(NULL,"NFactor"))
      m_pVertexConstants->SetVector(m_pDevice,handle,(D3DXVECTOR4*)&v4NFactor);

    if(handle = m_pVertexConstants->GetConstantByName(NULL,"Ambient"))
      m_pVertexConstants->SetVector(m_pDevice,handle,(D3DXVECTOR4*)&v4Ambient);

    if(handle = m_pVertexConstants->GetConstantByName(NULL,"MatDiff"))
      m_pVertexConstants->SetVector(m_pDevice,handle,(D3DXVECTOR4*)&v4MatDiffuse);

      // lights

    if(handle = m_pVertexConstants->GetConstantByName(NULL,"LightDiff1"))
      m_pVertexConstants->SetVector(m_pDevice,handle,(D3DXVECTOR4*)&v4LightDiffuse1);

    if(handle = m_pVertexConstants->GetConstantByName(NULL,"LightDir1"))
      m_pVertexConstants->SetVector(m_pDevice,handle,(D3DXVECTOR4*)&v4LightDir1);

    if(handle = m_pVertexConstants->GetConstantByName(NULL,"LightDiff2"))
      m_pVertexConstants->SetVector(m_pDevice,handle,(D3DXVECTOR4*)&v4LightDiffuse2);

    if(handle = m_pVertexConstants->GetConstantByName(NULL,"LightDir2"))
      m_pVertexConstants->SetVector(m_pDevice,handle,(D3DXVECTOR4*)&v4LightDir2);

      // pshader texture

    if(m_pPixelConstants && (handle = m_pPixelConstants->GetConstantByName(NULL,"baseTex")))
      {
      D3DXCONSTANT_DESC constDesc;
      UINT count = 1;
      m_pPixelConstants->GetConstantDesc(handle,&constDesc,&count);

      if(constDesc.RegisterSet == D3DXRS_SAMPLER)
        TextureIndex = constDesc.RegisterIndex;
      }
    }

  // draw

  m_pDevice->SetVertexDeclaration(m_pVertexDeclaration);
  m_pDevice->SetVertexShader     (m_pVertexShader);
  m_pDevice->SetPixelShader      (m_pPixelShader);

  m_pDevice->SetStreamSource     (0,m_pVB,0,sizeof(VERTEX_SIMPLE));
  m_pDevice->SetIndices          (  m_pIB);
  m_pDevice->SetTexture          (TextureIndex,m_pTexFront);

  D3DPRIMITIVETYPE Type = D3DPT_TRIANGLELIST;

  HRESULT hrErr = m_pDevice->DrawIndexedPrimitive(Type,0,0,4,0,2);    // 4 vtx, 2 tris
  if(FAILED(hrErr)) return hrErr;

  // 2-sided => render other side

  if(m_bo2Sided)
    {
    // ASM

    v4NFactor.Set(-1.f);
    m_pDevice->SetVertexShaderConstantF(14,(float*)&v4NFactor,1);

    // HLSL

    if(m_pVertexConstants)
      {
      D3DXHANDLE handle;
      if(handle = m_pVertexConstants->GetConstantByName(NULL,"NFactor"))
        m_pVertexConstants->SetVector(m_pDevice,handle,(D3DXVECTOR4*)&v4NFactor);
      }

    //draw

    m_pDevice->SetTexture(TextureIndex,m_pTexBack);

    m_pDevice->SetRenderState(D3DRS_CULLMODE,D3DCULL_CCW);
    HRESULT hrErr = m_pDevice->DrawIndexedPrimitive(Type,0,0,4,0,2);
    m_pDevice->SetRenderState(D3DRS_CULLMODE,D3DCULL_CW);
    if(FAILED(hrErr)) return hrErr;
    }

  return 0;
  }

//--------------------------------------------------------------------------------------------------------------------//
//                                                TRF                                                                 //
//--------------------------------------------------------------------------------------------------------------------//

//----------------------------------------------- SetModel2World -----------------------------------------------------//
// set model->world trf
// 26/06/02, Mythos
// in : model->world matrix
// out: OK?
//--------------------------------------------------------------------------------------------------------------------//

bool CRendererDX9::SetModel2World(const CMat4x4& m4Matrix)
  {
  if(!m_pDevice) return false;
  return SUCCEEDED(m_pDevice->SetTransform(D3DTS_WORLD,(D3DMATRIX*)&m4Matrix));
  }

//----------------------------------------------- SetWorld2View ------------------------------------------------------//
// set world->view trf
// 26/06/02, Mythos
// in : world->view matrix
// out: OK?
//--------------------------------------------------------------------------------------------------------------------//

bool CRendererDX9::SetWorld2View(const CMat4x4& m4Matrix)
  {
  if(!m_pDevice) return false;
  return SUCCEEDED(m_pDevice->SetTransform(D3DTS_VIEW,(D3DMATRIX*)&m4Matrix));
  }

//----------------------------------------------- SetPerspective -----------------------------------------------------//
// set perspective projection
// 26/06/02, Mythos
// in : fov,near & far distances,screen ratio (=4/3)
// out: OK?
//--------------------------------------------------------------------------------------------------------------------//

bool CRendererDX9::SetPerspective(const float fFov,const float fNear,const float fFar,const float fRatio)
  {
  if(!m_pDevice) return false;

  float fRight = (float)tan(fFov*0.5f)*fNear;
  float fTop   = fRight/fRatio;

  D3DXMATRIX m4Proj;
  D3DXMatrixPerspectiveRH(&m4Proj,fRight*2.f,fTop*2.f,fNear,fFar);
  return SUCCEEDED(m_pDevice->SetTransform(D3DTS_PROJECTION,(D3DMATRIX*)&m4Proj));
  }

//----------------------------------------------- SetOrtho -----------------------------------------------------------//
// set orthographic projection
// 26/06/02, Mythos
// in : volume extents
// out: OK?
//--------------------------------------------------------------------------------------------------------------------//

bool CRendererDX9::SetOrtho(const CVect3D& v3Min,const CVect3D& v3Max)
  {
  if(!m_pDevice) return false;

  D3DXMATRIX m4Proj;
  D3DXMatrixOrthoOffCenterRH(&m4Proj,v3Min.x,v3Max.x,v3Min.y,v3Max.y,v3Min.z,v3Max.z);
  return SUCCEEDED(m_pDevice->SetTransform(D3DTS_PROJECTION,(D3DMATRIX*)&m4Proj));
  }

//----------------------------------------------- SetProjection ------------------------------------------------------//
// set projection
// 15/03/03, Mythos
// in : proj matrix
// out: OK?
//--------------------------------------------------------------------------------------------------------------------//

bool CRendererDX9::SetProjection(const CMat4x4& m4Proj)
  {
  if(!m_pDevice) return false;
  return SUCCEEDED(m_pDevice->SetTransform(D3DTS_PROJECTION,(D3DMATRIX*)&m4Proj));
  }

//--------------------------------------------------------------------------------------------------------------------//

} // namespace
