//--------------------------------------------------------------------------------------------------------------------//
//                                                DX9 RENDERER CLASS                                                  //
//--------------------------------------------------------------------------------------------------------------------//

#ifndef   _MYTHOS_RENDERERDX9_H_
#define   _MYTHOS_RENDERERDX9_H_
#pragma    once

//----------------------------------------------- INCLUDES -----------------------------------------------------------//

#define   D3D_OVERLOADS
#include  <d3d9.h>
#include  <D3dx9core.h>
#include  <dxerr9.h>

#include  "Misc/BasicTypes.h"
#include  "Misc/Defines.h"
#include  "Misc/VectMat.h"

//----------------------------------------------- CLASSES ------------------------------------------------------------//

//--------------------------------------------------------------------------------------------------------------------//
//                                                CRendererDX9                                                        //
//--------------------------------------------------------------------------------------------------------------------//

namespace Mythos
{
  class CRendererDX9
  {
  // public enums

  public:

    typedef enum
      {
      _WINDOWED_              = 1 << 0,
      _ZBUFFER_               = 1 << 1,
      } RENDERER_CREATE;

    typedef enum
      {
      _CLEAR_BACK_            = 1 << 0,
      _CLEAR_DEPTH_           = 1 << 1,
      _CLEAR_STENCIL_         = 1 << 2,

      _CLEAR_BACKDEPTH_       = _CLEAR_BACK_ | _CLEAR_DEPTH_,
      _CLEAR_ALL_             = 0xffffffff,
      } RENDERER_CLEAR;

  // structs

    typedef struct VERTEX_SIMPLE
      {
      CVect3D m_XYZ;
      CVect3D m_Normal;
      CVect2D m_Tex;
      VERTEX_SIMPLE(CVect3D& XYZ,CVect3D& Normal,CVect2D Tex) { m_XYZ = XYZ; m_Normal = Normal; m_Tex = Tex; }
      VERTEX_SIMPLE() {}
      } VERTEX_SIMPLE;

  // public methods

  public:

    // constructors & destructor

                              CRendererDX9        (void);
    virtual                  ~CRendererDX9        (void);

    // basic ops

    virtual bool              Create              (const HWND hWnd,const u32 u32Flags=_WINDOWED_|_ZBUFFER_,
                                                   const u32 u32Width=0,const u32 u32Height=0);
    virtual void              Destroy             (void);

    virtual bool              BeginScene          (void);
    virtual bool              EndScene            (void);
    virtual bool              Clear               (const u32 u32Flags=_CLEAR_BACKDEPTH_);
    virtual bool              Swap                (const RECT* pSrcRect=NULL,const RECT* pDestRect=NULL,const u32 u32Vbls=0);

    virtual bool              SetWireframe        (const bool boON);
    virtual void              Set2Sided           (const bool boON);

    // draw

    virtual u32               Draw                (void);

    // trf

    virtual bool              SetModel2World      (const CMat4x4& m4Matrix);
    virtual bool              SetWorld2View       (const CMat4x4& m4Matrix);
    virtual bool              SetPerspective      (const float fFov,const float fNear,const float fFar,const float fRatio=4.f/3.f);
    virtual bool              SetOrtho            (const CVect3D& v3Min,const CVect3D& v3Max);
    virtual bool              SetProjection       (const CMat4x4& m4Proj);

  // protected methods

  protected:

    void                      InitStates          (void);
    bool                      InitScene           (void);

  // protected data

  protected:

    bool                      m_bo2Sided;

    // DX9 access

    LPDIRECT3DDEVICE9         m_pDevice;
    LPDIRECT3D9               m_pD3D;

    // IB/VB

    LPDIRECT3DINDEXBUFFER9    m_pIB;
    LPDIRECT3DVERTEXBUFFER9   m_pVB;

    // vertex shader

    LPDIRECT3DVERTEXSHADER9   m_pVertexShader;
    LPDIRECT3DVERTEXDECLARATION9 m_pVertexDeclaration;
    LPD3DXCONSTANTTABLE       m_pVertexConstants;

    LPDIRECT3DPIXELSHADER9    m_pPixelShader;
    LPD3DXCONSTANTTABLE       m_pPixelConstants;

    // textures

    LPDIRECT3DTEXTURE9        m_pTexFront;
    LPDIRECT3DTEXTURE9        m_pTexBack;
  };
}

//----------------------------------------------- INLINES ------------------------------------------------------------//

#ifndef   _DEBUG
#include  "RendererDX9.inl"
#endif

//--------------------------------------------------------------------------------------------------------------------//

#endif // _MYTHOS_RENDERERDX9_H_
