vs.1.1

//----------------------------------------------------------------------------------------
// vertex inputs
//----------------------------------------------------------------------------------------

#define iPos        v0                                      // vertex position
#define iNormal     v1                                      // vertex normal
#define iTex0       v2                                      // base texture coordinates

dcl_position        iPos
dcl_normal          iNormal
dcl_texcoord0       iTex0

//----------------------------------------------------------------------------------------
// constants
//----------------------------------------------------------------------------------------

def                 c0, 0, 0, 0, 0
#define Zero        c0                                      // c0       : 0;0;0;0

#define Matrix      c10                                     // c10-c13  : matrix
#define NFactor     c14                                     // c14      : normal factor (-1 or +1)
#define Ambient     c15                                     // c15      : global ambient * material ambient
#define MatDiff     c16                                     // c16      : material diffuse color
#define MatAlpha    c16.w

#define LightDiff1  c20                                     // c20      : light1 diffuse color
#define LightDir1   c21                                     // c21      : light1 dir in model space

#define LightDiff2  c30                                     // c30      : light2 diffuse color
#define LightDir2   c31                                     // c31      : light2 dir in model space

//----------------------------------------------------------------------------------------
// code
//----------------------------------------------------------------------------------------

m4x4                oPos, iPos, Matrix                      // transform position
mul                 r2, iNormal, NFactor                    // N or -N
mov                 oT0.xy, iTex0                           // copy tex coords

// directional light 1

dp3                 r0, r2, -LightDir1                      // N * -LightDir1
max                 r0, r0, Zero                            // clamp to [0;1]
mul                 r1, r0.x, MatDiff                       //   * material diffuse
mul                 r1, r1, LightDiff1                      //   * light1   diffuse

// directional light 2

dp3                 r0, r2, -LightDir2                      // N * -LightDir2
max                 r0, r0, Zero                            // clamp to [0;1]
mul                 r3, r0.x, MatDiff                       //   * material diffuse
mul                 r3, r3, LightDiff2                      //   * light2   diffuse

// final color

add                 r0, r1, r3
add                 oD0, r0, Ambient                        //   + ambient
mov                 oD0.w, MatAlpha                         // preserve alpha
